#ifdef UNITTEST5
#define NWAYCACHE_SUPPORT_STRING
#include <string>
#endif
#ifdef UNITTEST9
#include <pthread.h>
#endif

#include "nwaycache.h"
#include <assert.h>
using namespace NWayCache;

#ifdef UNITTEST1
//Verify utterly simple functionality
int main(void) {
	Cache<unsigned,unsigned,4> cache(10,300);
	
	cache.insert(17,42);
	cache.insert(18,43);
	cache.insert(19,44);
	
	assert(cache.exists(17));
	assert(cache.exists(18));
	assert(cache.exists(19));
}
#endif

#ifdef UNITTEST2
//Verify that a bit more
class Foo {
	int x;
public:
	Foo() : x(42) {}
	~Foo() { assert(x==42); x=0; }
};

int main(void) {
	Cache<int,Foo,4> cache(10,300);
	for(int i=0; i<1000; i++) {
		cache.insert(i,Foo());
	}
	for(int i=0; i<1000; i++) {
		cache.update(i,Foo());
	}
	assert(!cache.exists(1001));
}
#endif

#ifdef UNITTEST3
//Verify that the items creates constructed and destroyed properly
static int creates=0;
static int destroys=0;
class Foo {
public:
	Foo() { creates++; }
	Foo(const Foo&) { creates++; }
	~Foo() { destroys++; }
};

int main(void) {
	Cache<int,Foo,4> cache(10,300);
	for(int i=0; i<1000; i++) {
		cache.insert(i,Foo());
	}
	for(int i=0; i<1000; i++) {
		cache.update(i,Foo());
	}
	for(int i=0; i<1000; i++) {
		cache.insert(i,NotAvailable);
	}
	for(int i=0; i<1000; i++) {
		cache.update(i,NotAvailable);
	}
	for(int i=0; i<1000; i++) {
		cache.erase(i);
	}
	assert(creates>=1000);
	assert(creates==destroys);
}
#endif

#ifdef UNITTEST4
//Verify that basic functionalty works
int main(void) {
	Cache<int,int,4> cache(2,300);
	assert(!cache.exists(7));
	
	cache.insert(7,42);
	assert(cache.exists(7));
	
	assert(!cache.exists(1));
	assert(!cache.exists(2));
	assert(!cache.exists(3));
	assert(!cache.exists(4));
	assert(!cache.exists(5));
	assert(!cache.exists(6));
	assert(!cache.exists(8));
	
	cache.update(7,NotAvailable);
	assert(cache.exists(7));
	
	cache.erase(7);
	assert(!cache.exists(7));
	
	cache.insert(7,42);
	cache.update(7,43);
	assert(cache.exists(7));
	
	{
		Cache<int,int,4>::SlotReference_t sr;
		assert(cache.find(sr,7));
		assert(sr!=NotAvailable);
		assert(*sr == 43);
		*sr = 44;
		sr.update();
		assert(*sr == 44);
	}
	
	{
		Cache<int,int,4>::SlotReference_t sr;
		assert(cache.find(sr,7));
		assert(sr!=NotAvailable);
		assert(*sr == 44);
	}
	{
		Cache<int,int,4>::SlotReference_t sr;
		assert(cache.find(sr,7));
		sr.update(NotAvailable);
		assert(sr==NotAvailable);
	}
	{
		Cache<int,int,4>::SlotReference_t sr;
		assert(cache.find(sr,7));
		assert(sr==NotAvailable);
		sr.erase();
	}
	{
		Cache<int,int,4>::SlotReference_t sr;
		assert(!cache.find(sr,7));
	}
}
#endif

#ifdef UNITTEST5
//Verify that non-POD keys work
int main(void) {
	Cache<std::string,int,4> cache(2,300);
	assert(!cache.exists(std::string("a")));
	
	cache.insert(std::string("a"),42);
	cache.insert(std::string("b"),43);
	assert(cache.exists(std::string("a")));
	assert(cache.exists(std::string("b")));
}
#endif

#ifdef UNITTEST6
//Verify that statistics works
int main(void) {
	Cache<int,int,4> cache(25,300);
	unsigned long chp,chn,cm;
	for(int i=0; i<100; i++)
		cache.insert(i,i);
	for(int i=0; i<100; i++)
		cache.exists(i);
	cache.queryStatistics(chp,chn,cm);
	assert(chp==100);
	assert(cm==0);
	for(int i=100; i<150; i++)
		cache.insert(i,i);
	for(int i=0; i<100; i++)
		cache.exists(i);
	cache.queryStatistics(chp,chn,cm);
	assert(chp==150);
	assert(cm==50);
}
#endif

#ifdef UNITTEST7
//Verify that statistics works (more advanced)
#include <stdlib.h>
int main(void) {
	Cache<int,int,4> cache(250,300);
	
	for(int i=0; i<500; i++) {
		cache.insert(rand()%1000,i);
	}
	for(int i=0; i<500; i++) {
		cache.insert(rand()%1000,NotAvailable);
	}
	for(int i=0; i<1000; i++)
		cache.exists(rand()%1000);
	
	unsigned long chp,chn,cm;
	cache.queryStatistics(chp,chn,cm);
	assert(chp>0);
	assert(chn>0);
	assert(cm>0);
	assert(chp+chn+cm==1000);
}
#endif

#ifdef UNITTEST8
//Verify that the iterator works.
#include <stdlib.h>
int main(void) {
	Cache<int,int,4> cache(250,300);
	
	for(int i=0; i<500; i++) {
		cache.insert(rand()%1000,i);
	}
	for(int i=0; i<500; i++) {
		cache.insert(rand()%1000,NotAvailable);
	}
	
	size_t items_traversed=0;
	size_t available_items=0;
	for(Cache<int,int,4>::iterator iter=cache.begin();
	    iter!=cache.end();
	    ++iter)
	{
		items_traversed++;
		if(iter.available())
			available_items++;
	}
	assert(items_traversed<=1000);
	assert(available_items>0);
}
#endif

#ifdef UNITTEST9
//Verify that the cache and bucket locking works
#include <stdlib.h>
int main(void) {
	typedef Cache<int,int,4,GenericHasher<int>,MutexLockObject,MutexLockObject> Cache_t;
	Cache_t cache(250,300);
	
	for(int i=0; i<500; i++) {
		cache.insert(rand()%1000,i);
	}
	for(int i=0; i<500; i++) {
		cache.insert(rand()%1000,NotAvailable);
	}
	for(int i=0; i<500; i++)
		cache.exists(rand()%1000);
	for(int i=0; i<500; i++) {
		Cache_t::SlotReference_t sr;
		if(cache.find(sr,rand()%1000))
			cache.erase(sr);
	}
}
#endif

#ifdef SPEEDTEST
#define R 10000
int main(void) {
	Cache<int,int,4> cache(100,300);
	for(int r=0; r<R; r++) {
		for(int i=0; i<500; i++)
			cache.insert(i,i);
		for(int i=0; i<500; i++) {
			Cache<int,int,4>::SlotReference_t sr;
			cache.find(sr,i);
		}
		for(int i=0; i<500; i++)
			cache.update(i,i+1);
		for(int i=0; i<500; i++)
			cache.erase(i);
	}
}
#endif
